/*
 * Copyright 2017 aquenos GmbH.
 * All rights reserved.
 * 
 * Please refer to the file NOTICE.txt in the root of the source distribution or
 * in the META-INF directory of the binary archive for licensing information.
 */

package com.aquenos.epics.jackie.diirt.vtype;

import java.time.Instant;

import org.apache.commons.lang3.builder.EqualsBuilder;
import org.apache.commons.lang3.builder.HashCodeBuilder;
import org.apache.commons.lang3.builder.ToStringBuilder;
import org.diirt.vtype.Alarm;
import org.diirt.vtype.AlarmSeverity;
import org.diirt.vtype.Time;

import com.aquenos.epics.jackie.common.value.ChannelAccessTimeValue;

/**
 * Base class for all classes that represent VTypes backed by Channel Access
 * values from EPICS Jackie. It implements the methods related to the alarm
 * status and time-stamp.
 * 
 * @author Sebastian Marsching
 */
public abstract class JackieAlarmAndTime<TimeValueType extends ChannelAccessTimeValue<?>>
        implements Alarm, Time {

    private static final long OFFSET_EPICS_TO_UNIX_EPOCH = 631152000L;

    protected final Instant disconnectTime;
    protected final boolean scalar;
    protected final TimeValueType timeValue;

    /**
     * <p>
     * Initializes this value with the specified backing value. The
     * {@code timeValue} is used for the alarm status and time-stamp.
     * </p>
     * 
     * <p>
     * If {@code disconnectTime} is not <code>null</code>, it is used as the
     * time-stamp and the alarm severity is set to {@code UNDEFINED}.
     * </p>
     * 
     * <p>
     * This constructor is package-private because the classes implementing the
     * VTypes form a closed hierarchy.
     * </p>
     * 
     * @param timeValue
     *            value used for alarm status and time-stamp.
     * @param disconnectTime
     *            time at which the channel was disconnected. Usually this
     *            should be <code>null</code>. If it is not <code>null</code>,
     *            the alarm severity is set to {@code UNDEFINED} and this
     *            parameter is used as the time-stamp (instead of using the
     *            time-stamp from the {@code timeValue}.
     * @param scalar
     *            <code>true</code> if the child class represents a scalar,
     *            <code>false</code> if it represents an array.
     * @throws IllegalArgumentException
     *             if {@code scalar} is <code>true</code> and {@code timeValue}
     *             has more or less than one element.
     */
    JackieAlarmAndTime(TimeValueType timeValue, Instant disconnectTime,
            boolean scalar) {
        if (timeValue == null) {
            throw new NullPointerException();
        }
        if (scalar && timeValue.getValueSize() != 1) {
            throw new IllegalArgumentException(
                    "The time value must have exactly one element.");
        }
        this.disconnectTime = disconnectTime;
        this.scalar = scalar;
        this.timeValue = timeValue;
    }

    @Override
    public boolean equals(Object obj) {
        if (this == obj) {
            return true;
        }
        // We use the fact that our value types form a closed hierarchy and
        // check whether the actual type is the same. This simplifies the
        // implementation of equals(...) in derived types.
        if (obj == null || !(this.getClass().equals(obj.getClass()))) {
            return false;
        }
        JackieAlarmAndTime<?> other = (JackieAlarmAndTime<?>) obj;
        return new EqualsBuilder()
                .append(this.disconnectTime, other.disconnectTime)
                .append(this.scalar, other.scalar)
                .append(this.timeValue, other.timeValue).build();
    }

    @Override
    public String getAlarmName() {
        if (disconnectTime != null) {
            return "Disconnected";
        } else {
            return timeValue.getAlarmStatus().toString();
        }
    }

    @Override
    public AlarmSeverity getAlarmSeverity() {
        if (disconnectTime != null) {
            return AlarmSeverity.UNDEFINED;
        }
        switch (timeValue.getAlarmSeverity()) {
        case INVALID_ALARM:
            return AlarmSeverity.INVALID;
        case MAJOR_ALARM:
            return AlarmSeverity.MAJOR;
        case MINOR_ALARM:
            return AlarmSeverity.MINOR;
        case NO_ALARM:
            return AlarmSeverity.NONE;
        default:
            break;
        }
        return null;
    }

    @Override
    public Instant getTimestamp() {
        if (disconnectTime != null) {
            return disconnectTime;
        }
        // The Channel Access protocol uses unsigned 32-bit integers, so we have
        // to make sure that we treat the integers as unsigned..
        return Instant.ofEpochSecond(
                (timeValue.getTimeSeconds() & 0xffffffffL)
                        + OFFSET_EPICS_TO_UNIX_EPOCH,
                timeValue.getTimeNanoseconds() & 0xffffffffL);
    }

    @Override
    public Integer getTimeUserTag() {
        // Channel Access does not provide an extra time tag.
        return null;
    }

    @Override
    public int hashCode() {
        return new HashCodeBuilder(97, 31).append(disconnectTime).append(scalar)
                .append(timeValue).build();
    }

    @Override
    public boolean isTimeValid() {
        if (disconnectTime != null) {
            return true;
        }
        // EPICS Base considers a time-stamp invalid if both the seconds and the
        // nanoseconds field are null.
        return timeValue.getTimeSeconds() != 0
                || timeValue.getTimeNanoseconds() != 0;
    }

    @Override
    public String toString() {
        return new ToStringBuilder(this).append("disconnectTime")
                .append("scalar", scalar).append("timeValue", timeValue)
                .build();
    }

}
