/*
 * Copyright 2017 aquenos GmbH.
 * All rights reserved.
 * 
 * Please refer to the file NOTICE.txt in the root of the source distribution or
 * in the META-INF directory of the binary archive for licensing information.
 */

package com.aquenos.epics.jackie.diirt.vtype;

import java.time.Instant;
import java.util.List;

import org.diirt.vtype.VEnum;
import org.diirt.vtype.VTypeToString;

import com.aquenos.epics.jackie.common.value.ChannelAccessGraphicsEnum;
import com.aquenos.epics.jackie.common.value.ChannelAccessTimeEnum;

/**
 * {@link VEnum} implementation that is backed by a
 * {@link ChannelAccessGraphicsEnum} and {@link ChannelAccessTimeEnum}.
 * 
 * @author Sebastian Marsching
 */
public final class JackieVEnum extends JackieEnum implements VEnum {

    /**
     * <p>
     * Creates a value backed by the specified {@code graphicsValue} and
     * {@code timeValue}. The {@code graphicsValue} is used for the meta-data
     * (labels) and the {@code timeValue} is used for the alarm status,
     * time-stamp, and value. The {@code timeValue} must have exactly one
     * element.
     * </p>
     * 
     * <p>
     * If {@code disconnectTime} is not <code>null</code>, it is used as the
     * time-stamp and the alarm severity is set to {@code UNDEFINED}.
     * </p>
     * 
     * @param graphicsValue
     *            value used for meta-data (labels).
     * @param timeValue
     *            value used for alarm status, time-stamp, and value. It must
     *            have exactly one element.
     * @param disconnectTime
     *            time at which the channel was disconnected. Usually this
     *            should be <code>null</code>. If it is not <code>null</code>,
     *            the alarm severity is set to {@code UNDEFINED} and this
     *            parameter is used as the time-stamp (instead of using the
     *            time-stamp from the {@code timeValue}.
     * @throws IllegalArgumentException
     *             if {@code timeValue} has more or less than one element.
     */
    public JackieVEnum(ChannelAccessGraphicsEnum graphicsValue,
            ChannelAccessTimeEnum timeValue, Instant disconnectTime) {
        super(graphicsValue, timeValue, disconnectTime, true);
    }

    @Override
    public String getValue() {
        int index = getIndex();
        List<String> labels = getLabels();
        if (index < labels.size()) {
            return labels.get(index);
        } else {
            return Integer.toString(index);
        }
    }

    @Override
    public int getIndex() {
        // Enums are unsigned 16-bit integers, so we have to make sure they
        // are treated as unsigned.
        return timeValue.getValue().get(0) & 0xffff;
    }

    @Override
    public String toString() {
        return VTypeToString.toString(this);
    }

}
